#!/usr/bin/env python3
import numpy as np
import netCDF4 as nc

__version__ = '1.1.0'

class L3BinRead(object):
    """
    A Level 3 bin (L3b) file reader class

    This simple class provides methods for opening and reading OBPG L3b files

    """
    def __init__(self, *argv):
        self._ncid = None
        self._gid = None
        self._binList = None
        self._weights = None
        self._bins = None
        self.nobs = None
        self.nscenes = None
        self.resolution = None
        self.nrows = None
        self.productList = None
        self.products = {}

    # OPENING L3b
    def open(self,filename):
        '''open a netCDF level 3 bin file and fill out the class variables:
            _ncid: netCDF filehandle
            _gid: netCDF group ID for the 'level-3_binned_data' group
            _binList: BinList compound variable
            _weights: weights from the BinList
            bins: bin numbers from the BinList
            nobs: number of observations from the BinList
            nscenes: number of scenes from the BinList
            products: array of product variables stored in the bin file
            productList: array product names
        '''
        self._ncid = nc.Dataset(filename)
        self._gid = self._ncid.groups['level-3_binned_data']
        self._binList = self._gid.variables['BinList'][:]
        self._weights = self._binList['weights']
        self._bins = self._binList['bin_num']
        self.nobs = self._binList['nobs']
        self.nscenes = self._binList['nscenes']

        self.resolution = self._ncid.spatialResolution
        self.nrows = len(self._gid['BinIndex'].get_dims()[0])
        self.products = [self._gid.variables[product] for product in self._gid.variables.keys()  if product not in ['BinList', 'BinIndex']]
        self.productList = [product for product in self._gid.variables.keys()  if product not in ['BinList', 'BinIndex']]

    def close(self):
        '''close the netCDF filehandle'''
        self._ncid.close()

    def listProducts(self):
        '''return the products in the bin file'''
        return self.productList

    def consecutive(self, recordID, stepsize=1):
        '''return array of arrays with consecutive recordIDs'''
        return np.split(recordID, np.where(np.diff(recordID) != stepsize)[0]+1)

    def binNum2recordID(self, binnum):
        '''return the file bin index (recordID)(s) for bin number(s)'''
        return np.intersect1d(binnum, self._bins, assume_unique=True, return_indices=True)[2]

    def _getSum(self, prodID, recordID):
        '''return the sum squared for the product ID and bin of interest'''
        if len(recordID) < 10:
            return self.products[prodID][recordID]['sum']
        else:
            result = np.ones(len(recordID))
            start = 0
            for set in self.consecutive(recordID):
                if len(set) == 1:
                    result[start] = self.products[prodID][set[0]]['sum']
                    start += 1
                else:
                    end = len(set) + start
                    result[start:end] = self.products[prodID][set[0]:set[-1]+1]['sum']
                    start = end
            return result

    def _getSumSquared(self, prodID, recordID):
        '''return the sum squared for the product ID and bin of interest'''
        if len(recordID) < 10:
            return self.products[prodID][recordID]['sum_squared']
        else:
            result = np.ones(len(recordID))
            start = 0
            for set in self.consecutive(recordID):
                if len(set) == 1:
                    result[start] = self.products[prodID][set[0]]['sum_squared']
                    start += 1
                else:
                    end = len(set) + start
                    result[start:end] = self.products[prodID][set[0]:set[-1]+1]['sum_squared']
                    start = end
            return result

    def _getWeights(self, recordID):
        '''return the sum squared for the bin of interest'''
        if len(recordID) < 10:
            return self._weights[recordID]
        else:
            result = np.ones(len(recordID))
            start = 0
            for set in self.consecutive(recordID):
                if len(set) == 1:
                    result[start] = self._weights[set[0]]
                    start += 1
                else:
                    end = len(set) + start
                    result[start:end] = self._weights[set[0]:set[-1]+1]
                    start = end
            return result

    def get_nscenes(self, recordID):
        '''return the number of scenes for the bin of interest'''
        if len(recordID) < 10:
            return self.nscenes[recordID]
        else:
            result = np.ones(len(recordID))
            start = 0
            for set in self.consecutive(recordID):
                if len(set) == 1:
                    result[start] = self.nscenes[set[0]]
                    start += 1
                else:
                    end = len(set) + start
                    result[start:end] = self.nscenes[set[0]:set[-1]+1]
                    start = end
            return result

    def get_nobs(self, recordID):
        '''return the number of observations for the bin of interest'''
        if len(recordID) < 10:
            return self.nobs[recordID]
        else:
            result = np.ones(len(recordID))
            start = 0
            for set in self.consecutive(recordID):
                if len(set) == 1:
                    result[start] = self.nobs[set[0]]
                    start += 1
                else:
                    end = len(set) + start
                    result[start:end] = self.nobs[set[0]:set[-1]+1]
                    start = end
            return result

    def getMean(self, productName, recordID):
        '''calculate the mean(s) for a bin (or set of bins) for a product provided the product ID and recordID(s)'''
        prodID = self.productList.index(productName)
        return self._getSum(prodID, recordID) / self._getWeights(recordID)

    def getVariance(self, productName, recordID):
        '''calculate the variance for a product provided the product ID and bin of interest'''
        prodID = self.productList.index(productName)
        prodsum = self._getSum(prodID, recordID)
        prodsum_sq = self._getSumSquared(prodID, recordID)
        weights = self._getWeights(recordID)
        nscenes = self.get_nscenes(recordID)
        nonZeroID = np.asarray((np.square(weights) - nscenes) > 0).nonzero()[0]
        val = np.zeros(len(recordID))
        start = 0
        for set in self.consecutive(nonZeroID):
            if len(set) == 1:
                val[start] = (prodsum_sq[set[0]] / weights[set[0]]) - np.square(prodsum[set[0]] / weights[set[0]])
                val[start] *= np.square(weights[set[0]]) / (np.square(weights[set[0]]) - nscenes[set[0]])
                start += 1
            else:
                end = len(set) + start
                val[start:end] = (prodsum_sq[set[0]:set[-1]+1] / weights[set[0]:set[-1]+1]) - np.square(prodsum[set[0]:set[-1]+1] / weights[set[0]:set[-1]+1])
                val[start:end] *= np.square(weights[set[0]:set[-1]+1]) / (np.square(weights[set[0]:set[-1]+1]) - nscenes[set[0]:set[-1]+1])
                start = end

        return val

    def getStdev(self, productName, bin):
        '''calculate the standard deviation(s) for a bin (or set of bins) for a product provided the product ID and recordID(s)'''
        return np.sqrt(self.getVariance(productName, bin))
    
    def getAreaMean(self, productName, recordID):
        '''calculate the mean for a product provided the product ID and a set of bins (area of interest)'''
        return np.nanmean(self.getMean(productName, recordID))

    def getAreaMedian(self, productName, recordID):
        '''calculate the mean for a product provided the product ID and a set of bins (area of interest)'''
        return np.nanmedian(self.getMean(productName, recordID))

    def getAreaStdev(self, productName, recordID):
        '''calculate the standard deviation for a product provided the product ID and a set of bins (area of interest)'''
        return np.nanmean(self.getStdev(productName, recordID))

    def getAreaStats(self, productName, recordID):
        '''return the mean, median, and standard deviation of an area of interest'''
        areaBinMeans = self.getMean(productName, recordID)
        return np.nanmean(areaBinMeans), np.nanmedian(areaBinMeans), np.nanstd(areaBinMeans)
