#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Author: Brendan DeTracey
Code adapted from seadas v7.5:
build/src/libbin++/L3BinShape.cpp
build/src/libbin++/bin_io.cpp
/*
 The following functions are based on the pseudocode found in Appendix A of:

 Campbell, J.W., J.M. Blaisdell, and M. Darzi, 1995:
 Level-3 SeaWiFS Data Products: Spatial and Temporal Binning Algorithms.
 NASA Tech. Memo. 104566, Vol. 32,
 S.B. Hooker, E.R. Firestone, and J.G. Acker, Eds.,
 NASA Goddard Space Flight Center, Greenbelt, Maryland
 */

 Adapted and extended by Sean Bailey <sean.w.bailey@nasa.gov>
'''
import numpy as np

__version__ = '1.1.0'

_RESOLVE_NROWS = { '10D':(18, '10 degree'), '1D':(180, '1 degree'),
                   'HD':(360, 'half degree'), 'QD':(720, 'quarter degree'), 
                   '36':(2160//4, '36 km'), '18':(2160//2,'18.5 km'), 
                   '9':(2160, '9.2 km'), '4':(2160*2, '4.6 km'), 
                   '2':(2160*4,'2.3 km'), '1':(2160*8, '1.1 km'),
                   'H':(2160*16, '575 m'), 'Q':(2160*32, '288 m'),
                   'HQ':(2160*64, '144 m') }

_F64_HALF = np.float64(0.5)
_F64_90 = np.float64(90.)
_F64_180 = np.float64(180.)
_F64_360 = np.float64(360.)
_I32_0 = np.int32(0)
_I32_1 = np.int32(1)
_I32_2 = np.int32(2)
_SEAMLON = np.float64(-180)

class L3BinGrid(object):
    """
    A Level 3 bin (L3b) class

    This class provides methods for forward and inverse calculations of the 
    integerized sinusoidal grid bin numbers and latitudes and longitudes

    The class can be initialized with totalrows or the string "resolve" which
    will set total rows.  Default is to initialize to a 4.62km grid (resolve = "4")
    """
    def __init__(self,*argv,totalrows=4320,resolve=None):
        if resolve:
                self.totalrows = np.int32(_RESOLVE_NROWS[resolve][0])
        elif totalrows in [i[0] for i in list(_RESOLVE_NROWS.values())]:
            self.totalrows = totalrows
        else:
                self._usage()
        self.basebin = np.zeros(self.totalrows, dtype=np.int32)
        self.basebin[0] = _I32_1        
        row = np.arange(self.totalrows, dtype=np.int32)
        self.latbin = (row + _F64_HALF) * _F64_180 / self.totalrows - _F64_90
        self.numbin = np.int32(np.float64(_I32_2 * self.totalrows) * 
                      np.cos(np.deg2rad(self.latbin)) + _F64_HALF)
        self.basebin[1:] = np.cumsum(self.numbin[:-1]) + self.basebin[0]
        self.totalbins = self.basebin[self.totalrows - 1] + self.numbin[self.totalrows - 1] - 1


    def _usage(self):
        keys = _RESOLVE_NROWS.keys()
        print('The resolove option must be one of:' + str(keys))
        print('or the totalrows must be in: ' + ",".join([str(i[0]) for i in list(_RESOLVE_NROWS.values())]))

    def _constrain_bin(self, _bin):
        '''Constrain bin value to valid range'''
        _bin[_bin < 1] = _I32_1
        _bin[_bin > self.totalbins] = self.totalbins
        return _bin

    def _constrain_lat(self, lat):
        '''Constrain latitude values to valid range'''
        lat[lat > _F64_90] = _F64_90
        lat[lat < -_F64_90] = -_F64_90
        return lat

    def _constrain_lon(self, lon):
        '''Constrain longitude values to valid range'''
        idx = lon < -_F64_180
        counter = 0
        while np.count_nonzero(idx) > 0 & counter < 10:
            counter += 1
            lon[idx] += _F64_360
            idx = lon < -_F64_180
        idx = lon > _F64_180
        assert counter < 10
        counter = 0
        while np.count_nonzero(idx) > 0 & counter < 10:
            counter += 1
            lon[idx] -= _F64_360
            idx = lon > _F64_180
        assert counter < 10
        return lon

    def _constrain_row(self, row):
        '''Constrain row value to valid range'''
        row[row < _I32_0] = _I32_0
        row[row >= self.totalrows] = self.totalrows - _I32_1
        return row

    def _constrain_rowcol(self, row, col):
        '''Constrain row  and column values to valid range'''
        row[row < _I32_0] = _I32_0
        row[row >= self.totalrows] = self.totalrows - _I32_1
        col[col < _I32_0] = _I32_0
        idx = col >= self.numbin[row]
        col[idx] = self.numbin[row[idx]] - 1
        return row, col

    def bin2bounds(self, _bin):
        """
        Returns the bounding north, south, east and west values  corresponding to a given bin number.

            Parameters:
                    bin (int): L3 bin number

            Returns:
                    north (float): northernmost latitude
                    south (float): southernmost latitude
                    east (float): easternmost longitude
                    west (float): westernmost longitude
        """
        if np.isscalar(_bin):
            _bin = np.asarray([_bin], dtype=np.int32)
        else:
            _bin = np.asarray(_bin, dtype=np.int32)
        _bin = self._constrain_bin(_bin)
        row, col = self.bin2rowcol(_bin)
        north, south, east, west = self.rowcol2bounds(row, col)
        return north, south, east, west

    def bin2lonlat(self, _bin):
        """
        Returns the longitude and latitude  corresponding to a given bin number.

            Parameters:
                    bin (int): L3 bin number

            Returns:
                    lon (float): longitude 
                    lat (float): latitude
        """
        if np.isscalar(_bin):
            _bin = np.asarray([_bin], dtype=np.int32)
        else:
            _bin = np.asarray(_bin, dtype=np.int32)
        _bin = self._constrain_bin(_bin)
        row, col = self.bin2rowcol(_bin)
        lon, lat = self.rowcol2lonlat(row, col)            
        return lon, lat           
            
    def bin2row(self, _bin):
        """
        Returns the row corresponding to a bin number.

            Parameters:
                    _bin (int): L3 bin  

            Returns:
                    row (int): L3 row number
        """
        if np.isscalar(_bin):
            _bin = np.asarray([_bin], dtype=np.int32)
        else:
            _bin = np.asarray(_bin, dtype=np.int32)
        _bin = self._constrain_bin(_bin)
        row_mask = np.ones(_bin.shape, dtype=bool) # row mask set to True for rows found.
        rlow = np.zeros(_bin.shape, dtype=np.int32)
        rhigh = np.ones(_bin.shape, dtype=np.int32) * self.totalrows - _I32_1
        idx = _bin < _I32_1
        row_mask[idx] = False # South pole. Leave rlow as zero.
        idx = _bin >= self.totalbins
        row_mask[idx] = False
        rlow[idx] = self.totalrows - _I32_1 # North pole. Set to highest row.
        while np.count_nonzero(row_mask) > 0:
            rmid = (rlow + rhigh + _I32_1) // _I32_2
            idx = _bin < self.basebin[rmid]
            rhigh[idx] = rmid[idx] - _I32_1
            rlow[~idx] = rmid[~idx]
            row_mask[rlow == rhigh] = False
        return rlow
        
    def bin2rowcol(self, _bin):
        """
        Returns the row and column corresponding to a bin number.

            Parameters:
                    _bin (int): L3 bin  

            Returns:
                    row (int): L3 row number
                    col (int): L3 column number
        """
        if np.isscalar(_bin):
            _bin = np.asarray([_bin], dtype=np.int32)
        else:
            _bin = np.asarray(_bin, dtype=np.int32)
        _bin = self._constrain_bin(_bin)
        row = self.bin2row(_bin)
        col = _bin - self.basebin[row]
        return row, col

    def lat2row(self, lat):
        """
        Returns the bin row number corresponding to a given latitude.

            Parameters:
                    lat (float): latitude

            Returns:
                    row (int): L3 row number
        """
        if np.isscalar(lat):
            lat = np.asarray([lat], dtype=np.float64)
        else:
            lat = np.asarray(lat, dtype=np.float64)
        lat = self._constrain_lat(lat)
        row = np.int32((_F64_90 + lat) * self.totalrows / _F64_180)
        row = self._constrain_row(row)
        return row

    def lonlat2bin(self, lon, lat):
        """
        Returns the bin number corresponding to a given longitude and latitude.

            Parameters:
                    lon (float): longitude 
                    lat (float): latitude

            Returns:
                    bin (int): L3 bin number
        """
        if np.isscalar(lat):
            lat = np.asarray([lat], dtype=np.float64)
        else:
            lat = np.asarray(lat, dtype=np.float64)
        if np.isscalar(lon):
            lon = np.asarray([lon], dtype=np.float64)
        else:
            lon = np.asarray(lon, dtype=np.float64)
        assert (lat.shape == lon.shape) & (lat.size == lon.size)
        row = self.lat2row(lat)
        lon = self._constrain_lon(lon)
        col = np.int32(self.numbin[row] * (lon - _SEAMLON) / _F64_360)
        _bin = self.rowcol2bin(row, col)
        return _bin

    def lonlat2rowcol(self, lon, lat):
        """
        Returns the row and column corresponding to a given longitude and latitude.

            Parameters:
                    lon (float): longitude 
                    lat (float): latitude

            Returns:
                    row (int): L3 row number
                    col (int): L3 column number
        """
        if np.isscalar(lat):
            lat = np.asarray([lat], dtype=np.float64)
        else:
            lat = np.asarray(lat, dtype=np.float64)
        if np.isscalar(lon):
            lon = np.asarray([lon], dtype=np.float64)
        else:
            lon = np.asarray(lon, dtype=np.float64)
        assert (lat.shape == lon.shape) & (lat.size == lon.size)
        row = self.lat2row(lat)
        lon = self._constrain_lon(lon)
        col = np.int32(self.numbin[row] * (lon - _SEAMLON) / _F64_360)
        return row, col

    def rowcol2bin(self, row, col):
        """
        Returns the bin number corresponding to a given row and column.

            Parameters:
                    row (int): L3bin row 
                    col (int): L3bin column

            Returns:
                    bin (int): L3 bin number
        """
        if np.isscalar(row):
            row = np.asarray([row], dtype=np.int32)
        else:
            row = np.asarray(row, dtype=np.int32)
        if np.isscalar(col):
            col = np.asarray([col], dtype=np.int32)
        else:
            col = np.asarray(col, dtype=np.int32)
        assert (row.shape == col.shape) & (row.size == col.size)
        row, col = self._constrain_rowcol(row, col)
        _bin = self.basebin[row] + col
        return _bin

    def rowcol2bounds(self, row, col):
        """
        Returns the bounding north, south, east and west values  
        corresponding to a row and column.

            Parameters:
                    row (int): L3bin row 
                    col (int): L3bin column

            Returns:
                    north (float): northernmost latitude
                    south (float): southernmost latitude
                    east (float): easternmost longitude
                    west (float): westernmost longitude
        """
        if np.isscalar(row):
            row = np.asarray([row], dtype=np.int32)
        else:
            row = np.asarray(row, dtype=np.int32)
        if np.isscalar(col):
            col = np.asarray([col], dtype=np.int32)
        else:
            col = np.asarray(col, dtype=np.int32)
        assert (row.shape == col.shape) & (row.size == col.size)
        lon, lat = self.rowcol2lonlat(row, col)
        north = lat + _F64_90 / self.totalrows
        south = lat - _F64_90 / self.totalrows
        east = lon + _F64_180 / self.numbin[row]
        west = lon - _F64_180 / self.numbin[row]
        return north, south, east, west

    def rowcol2lonlat(self, row, col):
        """
        Returns the latitude and longitude number corresponding to a given row and column.

            Parameters:
                    row (int): L3bin row 
                    col (int): L3bin column

            Returns:
                    lon (float): longitude 
                    lat (float): latitude
        """
        if np.isscalar(row):
            row = np.asarray([row], dtype=np.int32)
        else:
            row = np.asarray(row, dtype=np.int32)
        if np.isscalar(col):
            col = np.asarray([col], dtype=np.int32)
        else:
            col = np.asarray(col, dtype=np.int32)
        assert (row.shape == col.shape) & (row.size == col.size)
        row, col = self._constrain_rowcol(row, col)
        lat = self.latbin[row]        
        lon = _F64_360 * (col + _F64_HALF) / self.numbin[row] + _SEAMLON
        return lon, lat

    def rowlon2bin(self, row, lon):
        """
        Returns the bin number corresponding to a given row and longitude.

            Parameters:
                    row (int): L3bin row 
                    col (int): L3bin column

            Returns:
                    bin (int): L3 bin number
        """    
        if np.isscalar(row):
            row = np.asarray([row], dtype=np.int32)
        else:
            row = np.asarray(row, dtype=np.int32)
        if np.isscalar(lon):
            lon = np.asarray([lon], dtype=np.float64)
        else:
            lon = np.asarray(lon, dtype=np.float64)
        assert (row.shape == lon.shape) & (row.size == lon.size)
        lon = self._constrain_lon(lon)
        col = np.int32((lon - _SEAMLON) * self.numbin[row] / _F64_360)
        idx = col >= self.numbin[row]
        while np.count_nonzero(idx) > 0:
            col[idx] = self.numbin[row][idx] - _I32_1
            idx = col >= self.numbin[row]
        _bin = self.basebin[row] + col
        return _bin
    
    def region2bins(self, north, south, west, east):
        """
        Returns the bin numbers for a bounding north, south, east and west box  

            Parameters:
                north (float): northernmost latitude
                south (float): southernmost latitude
                east (float): easternmost longitude
                west (float): westernmost longitude

            Returns:
                numpy array of bin numbers
        """
        minrow = self.lat2row(south)[0]
        maxrow = self.lat2row(north)[0]
        def findbins():
            for row in range(minrow, maxrow):
                for bin in range(self.rowlon2bin(row,west)[0],self.rowlon2bin(row,east)[0]):
                    yield bin
        return np.fromiter(findbins(), dtype=int)
